#!/usr/local/bin/perl
# 
# the functions in this script are called bei exim embedded perl
# see exim/conf
# 

# see http://www.exim.org/eximwiki/FAQ/Perl/Q1303
$SIG{__WARN__} = sub { Exim::log_write($_[0]) };

use strict;
use warnings;
use BerkeleyDB;

# syntax of accounts files
# <local-part> <a> <b> <c> <d> <list of forwarder/pipe/mbox>

# global settings
my $popdir = "/var/pop";

# ----------------------------------------------------------------------- 
# make exim use vdom for delivery
sub get_destinations {

    my ( $exists ) = @_;

    #if ( !defined($exists) ) {
    #	Exim::log_write("exists: undefined");
    #}
    #else {
    #	Exim::log_write("exists: $exists");
    #}
  
    #system("(date;/usr/bin/id) >> /tmp/exim-id.debug");

    my $lp = Exim::expand_string('$local_part') ||
	die "Could not expand exim variable 'local_part'";
    my $dom = Exim::expand_string('$domain') ||
	die "Could not expand exim variable 'domain'";
    my $spamstat = Exim::expand_string('$header_X-Spam-Status:');

    #Exim::log_write("spamstat: $spamstat");

    my $accfile = "$popdir/$dom/Accounts";
  
    # Accounts-file zum Lesen oeffnen
    if ( ! open (ACCFILE, "<$accfile")) {
	# TODO: case "smarthost", ohne Accounts-File
	die "Could not read '$accfile'\n";
    }

    my $line;
    my $forward;
    my $default="postmaster";
    my $postmaster = "postmaster\@$dom";
  
    # Zeile mit local_part suchen
    while ($line= <ACCFILE>){
  
  	chomp($line);
  
  	# my $line =
	#   "user WacUnugEy - - a otheruser\@ab.c /var/pop/mydom.ain/user1";
  
  	# print join(":", split(/\s+/, $line, 6), "\n");
  
  	(my @tmp) = split(/\s+/, $line, 6);

	# ignore case of alias entries
  	$tmp[0] = lc($tmp[0]);
  
  	# print "tmp0='$tmp[0]'   tmp5='$tmp[5]'\n";
  
  	if ($tmp[0] eq "mailer-daemon") {
	    $default="fail";
  	}
  
        if ($tmp[0] eq $lp) {
                $forward = "$popdir/$dom/$lp";
                if ( $spamstat =~ /^Yes,/ &&
                     -s "$popdir/$dom/Conf/$lp.spamtag" )
		{
		    $forward .= ".spam";
                }

                if (defined($tmp[5]) && $tmp[5] ne "" && $forward !~ /\.spam$/ ) {
                        $forward = $tmp[5];
                }
  
                last;   # user gefunden
        } 
    }
  
    # print "forward='$forward'\n";

    if (not $forward) {
	#Exim::log_write("exists: forward is not defined");
  	if ($default eq "fail")	
  	{
	    if ( defined($exists) ) {
		#Exim::log_write("exists: address $lp missing, return false");
		return('false');
	    }
  	    $forward=":fail: unknown local-part <$lp>";
  	}
	elsif ($default eq "defer")
	{
	    $forward=":defer: relaying to <$lp> prohibited by administrator";
	}
  	else
	{
	    if ( defined($exists) ) {
		#Exim::log_write("exists: address $lp missing, return false");
		return('false');
	    }
  	    $forward=$postmaster;
	}
    }
    else {
	#Exim::log_write("exists: forward is '$forward'");
    }

    # dont split forwards with :fail: :defer: or :blackhole:
    my $dstlist = "";
    if ($forward =~ /^:/ ) {
   	$dstlist = $forward;
    }
    else {
	# process destlist
	# if the user has himself on the list, add his mailbox file
	my (@dest) = split(/\s+/, $forward );
	my ($i);
	foreach $i ( 0..$#dest) {
	    $dest[$i] = lc($dest[$i]);
	    if ( $dest[$i] eq $lp.'@'.$dom ) {
		$dest[$i] = "$popdir/$dom/$lp";
	    }
	}
	$dstlist = join(",", @dest);
    }
    &logging("dstlist: '$dstlist'");
    if ( defined($exists) ) {
	#Exim::log_write("exists: address $lp exists, return true");
	return('true');
    }
    return ($dstlist);
} # end sub get_destinations


# ----------------------------------------------------------------------- 
# make exim use vdom for delivery
# special case: only if a local forward exists, return true on 'exists'
# otherwise, return false, so that the next router is tried
sub get_local_destinations {

    my ( $exists ) = @_;

    #if ( !defined($exists) ) {
    #	Exim::log_write("local-exists: undefined");
    #}
    #else {
    #	Exim::log_write("local-exists: $exists");
    #}
  
    #system("(date;/usr/bin/id) >> /tmp/exim-id.debug");

    my $lp = Exim::expand_string('$local_part') ||
	die "Could not expand exim variable 'local_part'";
    my $dom = Exim::expand_string('$domain') ||
	die "Could not expand exim variable 'domain'";
    my $spamstat = Exim::expand_string('$header_X-Spam-Status:');

    #Exim::log_write("spamstat: $spamstat");

    my $accfile = "$popdir/$dom/Accounts";
  
    # Accounts-file zum Lesen oeffnen
    if ( ! open (ACCFILE, "<$accfile")) {
	# TODO: case "smarthost", ohne Accounts-File
	die "Could not read '$accfile'\n";
    }

    my $line;
    my $forward;
    my $realforward;
    my $default="postmaster";
    my $postmaster = "postmaster\@$dom";

    # Zeile mit local_part suchen
    while ($line= <ACCFILE>){

  	chomp($line);

  	# my $line =
	#   "user WacUnugEy - - a otheruser\@ab.c /var/pop/mydom.ain/user1";

  	# print join(":", split(/\s+/, $line, 6), "\n");

  	(my @tmp) = split(/\s+/, $line, 6);

	# ignore case of alias entries
  	$tmp[0] = lc($tmp[0]);

  	# print "tmp0='$tmp[0]'   tmp5='$tmp[5]'\n";

  	if ($tmp[0] eq "mailer-daemon") {
	    $default="fail";
  	}

        if ($tmp[0] eq $lp) {
                $forward = "$popdir/$dom/$lp";
                if ( $spamstat =~ /^Yes,/ &&
                     -s "$popdir/$dom/Conf/$lp.spamtag"
		   ) {
		    $forward .= ".spam";
                }

                if (defined($tmp[5]) && $tmp[5] ne "" && $forward !~ /\.spam$/ ) {
		    $realforward = $tmp[5];
                        $forward = $tmp[5];
                }
  
                last;   # user gefunden
        } 
    }
  
    if (not $forward) {
	#Exim::log_write("local-exists: forward is not defined");
  	if ($default eq "fail")	
  	{
	    if ( defined($exists) ) {
		#Exim::log_write("local-exists: address $lp missing, return false");
		return('false');
	    }
  	    $forward=":fail: unknown local-part <$lp>";
  	}
  	elsif ($default eq "defer")
  	{
  	    $forward=":defer: relaying to <$lp> prohibited by administrator";
  	}
  	else
  	{
	    if ( defined($exists) ) {
		#Exim::log_write("local-exists: address $lp missing, return false");
		return('false');
	    }
  	    $forward=$postmaster;
  	}
    }
    else {
	#Exim::log_write("local-exists: forward is '$forward'");
    }

    # dont split forwards with :fail: :defer: or :blackhole:
    my $dstlist = "";
    if ($forward =~ /^:/ ) {
   	$dstlist = $forward;
    }
    else {
	# process destlist
	# if the user has himself on the list, add his mailbox file
	my (@dest) = split(/\s+/, $forward );
	my ($i);
	foreach $i ( 0..$#dest) {
	    $dest[$i] = lc($dest[$i]);
	    if ( $dest[$i] eq $lp.'@'.$dom ) {
		$dest[$i] = "$popdir/$dom/$lp";
	    }
	}
	$dstlist = join(",", @dest);
    }
    &logging("dstlist: '$dstlist'");
    if ( defined($exists) ) {
	if ( defined($realforward) && $realforward ne '' ) {
	    #Exim::log_write("local-exists: realforward $lp exists, return true");
	    return('true');
	}
	else {
	    #Exim::log_write("local-exists: realforward $lp missing, return false");
	    return('false');
	}
    }
    return ($dstlist);
} # end sub get_destinations


# ----------------------------------------------------------------------- 
# POP before SMTP
# get_relayhosts muss eine colon-separated list mit IPs zurueckgeben.
sub get_relayhosts
{
    my $authtimeout = 1800;
    my $filename = "/var/run/qpopper" ;
    use vars qw( $env $db %hash $ipword $timestamp );

    my $db = new BerkeleyDB::Hash
                -Filename => $filename,
                -Flags    => DB_RDONLY
        or die "Cannot read database $filename: $! $BerkeleyDB::Error\n" ;

    # get Sender IP

    my $sender_ip = Exim::expand_string('$sender_host_address') ;
    # IP also dot-separated-decimal	

    &logging( join(":", "sender_ip", $sender_ip ));

    # IP dot-separated-decimal in word umwandeln
    $ipword = pack("C4", split(/\./, $sender_ip));

    # get timestamp
    my $cursor = $db->db_cursor() ;

    # my $ret = $cursor->db_get($ipword, $timestamp);
    my $ret = $cursor->c_get($ipword, $timestamp, DB_SET);

    &logging( "after db_get");

    if ($ret == 0) { # key found
	&logging ( join(":", "timestamp", $timestamp));
    }
    else {
	&logging ("no timestamp found");
	return ("127.0.0.1");
    }

    undef $cursor ;
    undef $db ;

    if (time < ($authtimeout + $timestamp) ) {
	return ($sender_ip);
    }
    else {
        return ("127.0.0.1");
    }
}


# ----------------------------------------------------------------------- 
# SMTP-Auth
sub user_check() {

    my $nutzer = lc(shift(@_));
    my $pass = shift(@_);
    my $answer = "";
    my $cnt = 0;

    Exim::log_write("in user_check: nutzer: $nutzer pass $pass\n");

    my @user = split(/[\+\@]/,$nutzer);

    my($line);
    my(@line);
    my($user);
    my($domain);
    my($olduser);

    if ( !defined($user[1]) ) {
	# we need to search the accounts file for the domain
	Exim::log_write("smtpauth: domain missing, searching flat username space\n");
	my($accfile) = "/usr/local/etc/dovecot/accounts";
	my($res) = open (INP, $accfile );
	if ( !defined($res) ) {	# failed to open
	    $answer = "no";
	    Exim::log_write("smtpauth: can not open $accfile: $!\n");
	    return($answer);
	}
	while (<INP>) {
	    chomp;
	    $line=$_;
            next if ( $line =~ /^#/ );
            # logging("reading accounts: $line");
            @line = split(/\s+/,$line);
            if ( $#line > 1 && $user[0] eq $line[0] ) {
                $user = $line[1];
                $domain = $line[2];
                last;
            }
        }
        close(INP);

        if ( !defined($domain) ) {
	    Exim::log_write("smtpauth: user $user[0] not found in $accfile: $!\n");
	    $answer = "no";
            return($answer);
        }
        $olduser = $user[0];
        Exim::log_write("user $user[0] in domain $domain found for user $olduser in $accfile");

	$user[0] = $user;
	$user[1] = $domain;
    }

    my $accfile = $popdir."/".$user[1]."/Accounts";

    if ( ! open (FILE, $accfile )) {
	die "user_check: Could not open '$accfile'\n";
    }

    while (<FILE>) {

	my @zeile = split(/\s+/,$_);
	chomp $zeile[1];
	$cnt++;

        # Exim::log_write("in pass_by_user: $user[0] pass: $pass und $zeile[1] $zeile[0]\n");

	# TODO: Wenn SPA gefordert, und Passwort encrypted, dann: error
	# testcase: mit Outlook Express 6
	if ( $user[0] eq $zeile[0] && pwcmp($pass,$zeile[1],$cnt) ) {
	    $answer = "yes";
	    last;
	}
	else {
	    $answer = "no";
	}
    }
    close(FILE);
    return ($answer);
}


# ----------------------------------------------------------------------- 
# NTLM-Auth
# in: Benutzername out: Passwort || undef
sub pass_by_user() {

    my $nutzer = shift;
    my $answer = undef;

    my @user = split(/\+/,$nutzer);

    my $accfile = $popdir."/".$user[1]."/Accounts";

    if ( ! open (FILE, $accfile )) {
	die "user_check: Could not open '$accfile'\n";
    }

    while (<FILE>) {
	my @zeile = split(/\s+/,$_,3);

	if ($user[0] eq $zeile[0]) {
	    # Exim::log_write("in pass_by_user: $user[0] und $zeile[1]\n");
	    $answer = $zeile[1];
	    last;
	}
    }
    close(FILE);
    return $answer;
}

# Passwortvergleich
sub pwcmp {
    my($inpclear,$inppw,$cnt) = @_;
    my($j,$tmppw);

    if ( $inppw !~ /^\$/ ) {		# cleartext password
	logging("found clearpw style $inppw in line $cnt");
	if ( $inpclear eq $inppw ) {
	    return(1);
	}
	return(0);
    }
    elsif ( $inppw =~ /^\$0\$/ ) {	# oldstyle password, fixup
	( $j, $j, $tmppw ) = split(/\$/,$inppw);
	logging("found oldstyle pw $tmppw in line $cnt");
	if ( crypt($inpclear, $tmppw) eq $tmppw ) {
	    return(1);
	}
	return(0);
    }
    elsif ( $inppw =~ /^\$1\$/ ) {	# md5 password
	logging("found md5style pw $inppw in line $cnt");
	if ( crypt($inpclear, $inppw) eq $inppw ) {
	    return(1);
	}
	return(0);
    }
    else {
	logging("found unknown pw style $inppw in line $cnt");
    }
    return(0);
}


# ----------------------------------------------------------------------- 
sub logging {

#    open(OUTPUT,">>/tmp/eximtest") || die "Can't open >>/tmp/eximtest \n";
#    my ($s) = @_;
#    print OUTPUT "$s\n";
#    close(OUTPUT);

}

